(async function (Drupal, drupalSettings) {
  'use strict';

  const { terms, unacceptableElements, unacceptableContainers, highlightAll, mainContentSelector: configuredMainContentSelector } = drupalSettings?.keyValueGlossary;

  // If for whatever reason there are no terms at this stage, we've got nothing
  // to do.
  if (!terms || Object.keys(terms).length === 0) {
    return;
  }

  const searchRegexes = createSearchRegexes(Object.keys(terms));
  // Use the configured values or fall back to defaults if not available.
  const UNACCEPTABLE_ELEMENTS = unacceptableElements || [
    'script',
    'style',
    'a',
    'button',
    'h1',
    'h2',
    'h3',
    'h4',
    'h5',
    'h6',
  ];
  const UNACCEPTABLE_CONTAINERS = unacceptableContainers || [];
  const UNACCEPTABLE_SELECTORS = [...UNACCEPTABLE_ELEMENTS, ...UNACCEPTABLE_CONTAINERS].join(',');
  const mainContentSelector = configuredMainContentSelector || 'main';
  const template = document.getElementById('key-value-glossary-term-template');

  Drupal.behaviors.glossaries = {
    attach: async (context) => {
      // Scan page text for glossary terms and inject tooltips.
      const matches = findMatches(context);
      highlightMatches(matches);
    },
  };

  /**
   * Create a list of regexes for the search strings.
   *
   * @param {string[]} searchStrings
   *   The strings to search for.
   *
   * @return {Map}
   *   A map where string -> regex
   */
  function createSearchRegexes(searchStrings) {
    const searches = new Map();

    // Create one replacement function instead of creating it in the loop
    const escapeRegex = str => {
      let replaced = str;

      // Replace regex special characters with escaped versions.
      replaced = replaced.replace(/[.*+?^${}()|[\]\\]/g, '\\$&');

      // Replace any non-escaped, non alphanumeric characters with wildcards
      replaced = replaced.replace(/[^\w\s]/g, '.?');

      return replaced;
    };

    searchStrings.forEach(string => {
      const regex = new RegExp(`\\b${escapeRegex(string)}(?:es|s)?\\b`, highlightAll ? 'gi' : 'i');
      searches.set(string, regex);
    });

    return searches;
  }

  /**
   * Determine if a node should be accepted by the TreeWalker.
   *
   * This step filters out nodes before we look for matching terms.
   *
   * @param {Text} node
   *   The node to check.
   * @return {NodeFilter.FILTER_ACCEPT | NodeFilter.FILTER_REJECT}
   *   Whether the node should be accepted or rejected.
   */
  function shouldAcceptNode(node) {
    // Check text content first since it's the cheapest operation.
    if (node.textContent.trim() === '') {
      return NodeFilter.FILTER_REJECT;
    }

    // Cache the parent element reference.
    const parent = node.parentElement;
    if (!parent) {
      return NodeFilter.FILTER_REJECT;
    }

    // Do the closest() check.
    if (parent.closest(UNACCEPTABLE_SELECTORS)) {
      return NodeFilter.FILTER_REJECT;
    }

    // Only create range if we passed the other checks.
    const range = document.createRange();
    range.selectNode(node);
    const rect = range.getBoundingClientRect();
    range.detach();

    return (rect.width <= 1 || rect.height <= 1)
      ? NodeFilter.FILTER_REJECT
      : NodeFilter.FILTER_ACCEPT;
  }

  /**
   * Find all matches of the search strings in the given context.
   *
   * @param {Node} context
   *   The context to search in.
   * @param {Map<string, regex>} searchRegexes
   *   The strings to search for.
   *
   * @return {Object[]}
   *   A list of matches.
   */
  function findMatches(context) {
    const matches = [];
    const mainContent = document.querySelector(mainContentSelector);

    // Use cached mainContent
    if (!mainContent) {
      return matches;
    }

    // Scan the main content of the page on initial page load.
    // Scan the context (if in main content) on subsequent behavior runs.
    let scanRoot;
    if (mainContent.contains(context)) {
      scanRoot = context;
    }
    else if (context.contains(mainContent)) {
      scanRoot = mainContent;
    }
    else {
      // Exit if there's no relationship between context and main-content.
      return matches;
    }

    // Collect text nodes that pass shouldAcceptNode.
    const walker = document.createTreeWalker(
      scanRoot,
      NodeFilter.SHOW_TEXT,
      {
        acceptNode: shouldAcceptNode,
      },
    );

    // Iterate over nodes to find usages of terms.
    let node = walker.nextNode();
    while (node) {
      const text = node.textContent;

      // Quit looping over nodes if we've used all the terms.
      if (!searchRegexes.size) {
        break;
      }

      // Loop over unfound terms.
      for (const [searchString, searchRegex] of searchRegexes) {
        if (text.match(searchRegex)) {
          matches.push({
            node,
            searchString,
            searchRegex,
          });

          // Avoid searching for this term again.
          if (!highlightAll) {
            searchRegexes.delete(searchString);
          }
        }
      }
      node = walker.nextNode();
    }

    return matches;
  }

  /**
   * Create HTML content for a tooltip by combining definitions with glossary
   * source links.
   *
   * @param {Object.<string, string>} definitions
   *   Object mapping glossary UUIDs to definition text.
   *
   * @return {string}
   *   HTML string containing definition text and source citation links.
   */
  function createTooltipContent(definitions) {
    return Object.entries(definitions).map(([uuid, definition]) => {
      return definition;
    }).join('<hr/>');
  }

  /**
   * Generate a unique ID for a tooltip.
   *
   * @return {string}
   *   A unique ID for a tooltip.
   */
  function generateTooltipId() {
    return 'tooltip_' + Math.random().toString(36).substring(2, 11);
  }

  /**
   * Strips HTML tags from a string.
   *
   * @param {string} html
   *   The HTML string to strip tags from.
   *
   * @return {string}
   *   The text without HTML tags.
   */
  function stripHtml(html) {
    const div = document.createElement('div');
    div.innerHTML = html;
    return div.textContent || div.innerText || '';
  }

  /**
   * Create the markup for a tooltip.
   *
   * @param {string} term
   *   The term to display in the tooltip.
   * @param {string} definition
   *   The definition of the tooltip.
   *
   * @return {Element}
   *   The markup for a tooltip.
   */
  function createTooltip(term, definition) {
    const tooltipId = generateTooltipId();
    const element = template.content.firstElementChild.cloneNode(true);
    const plainTextDefinition = template.dataset.plainTextDefinition ?? 'true';
    const elementHtml = element.outerHTML;
    const populatedHtml = elementHtml
      .replace(/\$\{id}/g, tooltipId)
      .replace(/\$\{term}/g, term)
      .replace(/\$\{definition}/g, plainTextDefinition === 'true' ? stripHtml(definition) : definition);
    const tempDiv = document.createElement('div');
    tempDiv.innerHTML = populatedHtml;
    return tempDiv.firstElementChild;
  }

  /**
   * Highlight matches in the text.
   *
   * @param {Array<{node: Node, searchRegex: RegExp, searchString: string}>}
   *   matches - Array of matches containing text nodes, search regex and
   *   search string
   *
   * @return {void}
   */
  function highlightMatches(matches) {
    // Group matches by their parent node to process them together.
    const matchesByNode = new Map();

    matches.forEach(match => {
      if (!matchesByNode.has(match.node)) {
        matchesByNode.set(match.node, []);
      }
      matchesByNode.get(match.node).push(match);
    });

    matchesByNode.forEach((nodeMatches, node) => {
      const parent = node.parentElement;
      if (!parent) {
        return;
      }

      const originalText = node.textContent;

      // Create a set to track processed search strings for THIS node only.
      const processedSearchStrings = new Set();

      // Find all matches and their positions first.
      const matchPositions = [];

      nodeMatches.forEach(({ searchRegex, searchString }) => {
        // Skip if this search string has already been processed for this node.
        if (processedSearchStrings.has(searchString)) {
          return;
        }

        // Create a global version of the regex to find ALL matches.
        let match;

        // Find ALL matches in this text node.
        while ((match = searchRegex.exec(originalText)) !== null) {
          const newMatch = {
            start: match.index,
            end: match.index + match[0].length,
            matchText: match[0],
            searchString: searchString,
          };

          // Check for overlaps with existing matches.
          let hasOverlap = false;
          for (let i = 0; i < matchPositions.length; i++) {
            const existingMatch = matchPositions[i];

            // Check if matches overlap.
            const overlaps = (
              (newMatch.start >= existingMatch.start && newMatch.start <= existingMatch.end) ||
              (newMatch.end >= existingMatch.start && newMatch.end <= existingMatch.end) ||
              (newMatch.start <= existingMatch.start && newMatch.end >= existingMatch.end)
            );

            if (overlaps) {
              hasOverlap = true;
              // If the new match is longer, replace the existing one.
              if (newMatch.matchText.length > existingMatch.matchText.length) {
                matchPositions[i] = newMatch;
              }
              break;
            }
          }

          // If no overlap was found, add the new match.
          if (!hasOverlap) {
            matchPositions.push(newMatch);
          }

          // If we're not going to highlight all terms, just the first instance,
          // we're done here.
          if (!highlightAll) {
            break;
          }

          // Prevent infinite loop if regex doesn't advance.
          if (match.index === searchRegex.lastIndex) {
            searchRegex.lastIndex++;
          }
        }

        // Mark this search string as processed for this node only.
        processedSearchStrings.add(searchString);
      });

      // Sort match positions by their start index (descending) to process from
      // right to left.
      matchPositions.sort((a, b) => b.start - a.start);

      // Process matches from right to left to avoid position shifts.
      let currentText = originalText;
      let currentNode = node;

      matchPositions.forEach(({ start, end, matchText, searchString }) => {
        // Create text nodes for before and after the match
        const beforeText = document.createTextNode(currentText.substring(0, start));
        const afterText = document.createTextNode(currentText.substring(end));

        // Create the tooltip
        const definition = createTooltipContent(terms[searchString]);
        const tooltip = createTooltip(matchText, definition);

        // Replace the current text node
        parent.insertBefore(beforeText, currentNode);
        parent.insertBefore(tooltip, currentNode);
        parent.insertBefore(afterText, currentNode);
        parent.removeChild(currentNode);

        // Update for next iteration - we're working backwards through the
        // string.
        currentText = currentText.substring(0, start);
        currentNode = beforeText;
      });
    });
  }

  // Adjusts the position of the popup to be in the viewport, if needed.
  document.addEventListener("click", function (event) {
    const popoverTrigger = event.target.closest(".key-value-glossary-popover-button");
    if (popoverTrigger) {
      const popoverId = popoverTrigger.getAttribute('popovertarget');
      const popoverRoot = document.getElementById(popoverId);

      // Calculate the position of the popover.
      const position = popoverTrigger.getBoundingClientRect();

      popoverRoot.style.top = `${position.bottom + window.scrollY}px`;
      if (position.left < 0)
        popoverRoot.style.left = `${position.left}px`;
      else {
        const overflow =  position.right - document.documentElement.clientWidth;
        if (window.innerWidth > 767)
          popoverRoot.style.left = `${position.right}px`;
        else {
          popoverRoot.style.right = `${overflow}px`;
        }
      }
    }
  });

  // Hides the popover if the close button is clicked within the popup.
  document.addEventListener("click", function (event) {
    if (event.target.classList.contains("key-value-glossary-popover-close")) {
      const popupRoot = event.target.closest(".key-value-glossary-popover");
      if (popupRoot.matches(":popover-open")) {
        popupRoot.hidePopover();
      }
    }
  });

})(Drupal, drupalSettings);
