<?php

namespace Drupal\key_value_glossary\Form;

use Drupal\Core\Url;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form for configuring glossary and reference fields.
 */
class KeyValueGlossaryFieldSettingsForm extends FormBase {

  /**
   * Constructs a KeyValueGlossaryFieldSettingsForm object.
   *
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entityFieldManager
   *   The entity field manager.
   * @param \Drupal\Core\Entity\EntityTypeBundleInfoInterface $entityTypeBundleInfo
   *   The entity type bundle info.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   */
  public function __construct(
    protected EntityFieldManagerInterface $entityFieldManager,
    protected EntityTypeBundleInfoInterface $entityTypeBundleInfo,
    protected EntityTypeManagerInterface $entityTypeManager,
  ) {}

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_field.manager'),
      $container->get('entity_type.bundle.info'),
      $container->get('entity_type.manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'key_value_glossary_field_settings_form';
  }

  /**
   * Gets the title for the field settings form.
   *
   * @param string $type
   *   The type of field ('glossary' or 'reference').
   * @param ?int $delta
   *   The index of the field configuration to delete.
   * @param ?string $op
   *   The operation being performed ('delete' or NULL).
   *
   * @return \Drupal\Core\StringTranslation\TranslatableMarkup
   *   The translated title for the form.
   */
  public function getTitle(string $type, ?int $delta = NULL, ?string $op = NULL): TranslatableMarkup {
    if ($op === 'delete' && $delta !== NULL) {
      return $this->t('Delete @type Field', ['@type' => $type === 'glossary' ? 'Glossary' : 'Reference']);
    }
    if (in_array($type, ['glossary', 'reference'])) {
      return $this->t('Add @type Field', ['@type' => $type === 'glossary' ? 'Glossary' : 'Reference']);
    }
    return $this->t('Key Value Glossary Field Settings');
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, $type = NULL, $delta = NULL, $op = NULL) {
    $config = $this->configFactory()->getEditable('key_value_glossary.settings');

    // Handle delete operation.
    if ($op === 'delete' && $delta !== NULL) {
      if ($type === 'glossary') {
        $glossary_fields = $config->get('glossary_fields') ?: [];
        if (isset($glossary_fields[$delta])) {
          unset($glossary_fields[$delta]);
          $config->set('glossary_fields', array_values($glossary_fields))->save();
          $this->messenger()->addStatus($this->t('Glossary field configuration has been deleted.'));
        }
      }
      elseif ($type === 'reference') {
        $reference_fields = $config->get('glossary_reference_fields') ?: [];
        if (isset($reference_fields[$delta])) {
          unset($reference_fields[$delta]);
          $config->set('glossary_reference_fields', array_values($reference_fields))->save();
          $this->messenger()->addStatus($this->t('Reference field configuration has been deleted.'));
        }
      }

      return $this->redirect('key_value_glossary.settings');
    }

    // Build the form for adding a new field configuration.
    $form['type'] = [
      '#type' => 'value',
      '#value' => $type,
    ];

    $form['entity_type'] = [
      '#type' => 'select',
      '#title' => $this->t('Entity Type'),
      '#options' => $this->getEntityTypeOptions($type),
      '#default_value' => 'node',
      '#required' => TRUE,
      '#ajax' => [
        'callback' => '::updateBundleOptions',
        'wrapper' => 'bundle-wrapper',
      ],
    ];

    $form['bundle_wrapper'] = [
      '#type' => 'container',
      '#attributes' => ['id' => 'bundle-wrapper'],
    ];

    $entity_type = $form_state->getValue('entity_type') ?? 'node';
    if ($entity_type) {
      $form['bundle_wrapper']['bundle'] = [
        '#type' => 'select',
        '#title' => $this->t('Bundle'),
        '#options' => $this->getBundleOptions($entity_type),
        '#required' => TRUE,
        '#ajax' => [
          'callback' => '::updateFieldOptions',
          'wrapper' => 'field-wrapper',
        ],
      ];
    }
    else {
      $form['bundle_wrapper']['bundle'] = [
        '#type' => 'select',
        '#title' => $this->t('Bundle'),
        '#options' => [],
        '#disabled' => TRUE,
      ];
    }

    $form['field_wrapper'] = [
      '#type' => 'container',
      '#attributes' => ['id' => 'field-wrapper'],
    ];

    $entity_type = $form_state->getValue('entity_type');
    $bundle = $form_state->getValue('bundle');

    if ($entity_type && $bundle) {
      $field_options = [];

      if ($type === 'glossary') {
        $field_options = $this->getKeyValueFieldOptions($entity_type, $bundle);
      }
      elseif ($type === 'reference') {
        $field_options = $this->getEntityReferenceFieldOptions($entity_type, $bundle);
      }

      $form['field_wrapper']['field_name'] = [
        '#type' => 'select',
        '#title' => $this->t('Field'),
        '#options' => $field_options,
        '#required' => TRUE,
      ];
    }
    else {
      $form['field_wrapper']['field_name'] = [
        '#type' => 'select',
        '#title' => $this->t('Field'),
        '#options' => [],
        '#disabled' => TRUE,
      ];
    }

    $form['actions'] = [
      '#type' => 'actions',
    ];

    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Save'),
    ];

    $form['actions']['cancel'] = [
      '#type' => 'link',
      '#title' => $this->t('Cancel'),
      '#url' => Url::fromRoute('key_value_glossary.settings'),
      '#attributes' => [
        'class' => ['button'],
      ],
    ];

    return $form;
  }

  /**
   * Ajax callback to update bundle options.
   */
  public function updateBundleOptions(array &$form, FormStateInterface $form_state) {
    return $form['bundle_wrapper'];
  }

  /**
   * Ajax callback to update field options.
   */
  public function updateFieldOptions(array &$form, FormStateInterface $form_state) {
    return $form['field_wrapper'];
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $config = $this->configFactory->getEditable('key_value_glossary.settings');
    $type = $form_state->getValue('type');
    $entity_type = $form_state->getValue('entity_type');
    $bundle = $form_state->getValue('bundle');
    $field_name = $form_state->getValue('field_name');

    $field_config = [
      'entity_type' => $entity_type,
      'bundle' => $bundle,
      'field_name' => $field_name,
    ];

    if ($type === 'glossary') {
      $glossary_fields = $config->get('glossary_fields') ?: [];
      $glossary_fields[] = $field_config;
      $config->set('glossary_fields', $glossary_fields)->save();
      $this->messenger()->addStatus($this->t('Glossary field configuration has been saved.'));
    }
    elseif ($type === 'reference') {
      $reference_fields = $config->get('glossary_reference_fields') ?: [];
      $reference_fields[] = $field_config;
      $config->set('glossary_reference_fields', $reference_fields)->save();
      $this->messenger()->addStatus($this->t('Reference field configuration has been saved.'));
    }

    $form_state->setRedirect('key_value_glossary.settings');
  }

  /**
   * Get entity type options.
   *
   * @param ?string $type
   *   The type of field ('glossary' or 'reference').
   *
   * @return array
   *   An array of entity type options.
   */
  protected function getEntityTypeOptions(?string $type) {
    $options = [];

    $entity_types = $this->entityTypeManager->getDefinitions();
    foreach ($entity_types as $entity_type_id => $entity_type) {
      if ($type === 'reference' && !$entity_type->hasLinkTemplate('canonical')) {
        continue;
      }
      if ($entity_type->entityClassImplements('\Drupal\Core\Entity\ContentEntityInterface')) {
        $options[$entity_type_id] = $entity_type->getLabel();
      }
    }

    return $options;
  }

  /**
   * Get bundle options for an entity type.
   *
   * @param string $entity_type_id
   *   The entity type ID.
   *
   * @return array
   *   An array of bundle options.
   */
  protected function getBundleOptions($entity_type_id) {
    $options = [];

    $bundles = $this->entityTypeBundleInfo->getBundleInfo($entity_type_id);
    foreach ($bundles as $bundle_id => $bundle_info) {
      $options[$bundle_id] = $bundle_info['label'];
    }

    return $options;
  }

  /**
   * Get Key / Value field options for an entity type and bundle.
   *
   * @param string $entity_type_id
   *   The entity type ID.
   * @param string $bundle
   *   The bundle ID.
   *
   * @return array
   *   An array of field options.
   */
  protected function getKeyValueFieldOptions($entity_type_id, $bundle) {
    $options = [];

    $fields = $this->entityFieldManager->getFieldDefinitions($entity_type_id, $bundle);
    foreach ($fields as $field_name => $field_definition) {
      if (in_array($field_definition->getType(), ['key_value_long', 'key_value'])) {
        $options[$field_name] = $field_definition->getLabel();
      }
    }

    return $options;
  }

  /**
   * Get entity_reference field options for an entity type and bundle.
   *
   * @param string $entity_type_id
   *   The entity type ID.
   * @param string $bundle
   *   The bundle ID.
   *
   * @return array
   *   An array of field options.
   */
  protected function getEntityReferenceFieldOptions($entity_type_id, $bundle) {
    $options = [];

    $fields = $this->entityFieldManager->getFieldDefinitions($entity_type_id, $bundle);
    foreach ($fields as $field_name => $field_definition) {
      if ($field_definition->getType() === 'entity_reference') {
        $options[$field_name] = $field_definition->getLabel();
      }
    }

    return $options;
  }

}
