<?php

namespace Drupal\key_value_glossary\Form;

use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Configure Key Value Glossary settings.
 */
class KeyValueGlossarySettingsForm extends ConfigFormBase {

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * The entity type bundle info.
   *
   * @var \Drupal\Core\Entity\EntityTypeBundleInfoInterface
   */
  protected $entityTypeBundleInfo;

  /**
   * Module handler service.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $instance = parent::create($container);
    $instance->entityFieldManager = $container->get('entity_field.manager');
    $instance->entityTypeBundleInfo = $container->get('entity_type.bundle.info');
    $instance->moduleHandler = $container->get('module_handler');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return [
      'key_value_glossary.settings',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'key_value_glossary_settings_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $config = $this->config('key_value_glossary.settings');

    $form['description'] = [
      '#markup' => $this->t('Configure which fields should be used as glossaries and which fields should be used for referencing glossaries.'),
    ];

    $form['glossary_fields'] = [
      '#type' => 'details',
      '#title' => $this->t('Glossary fields'),
      '#description' => $this->t('Select which Key / Value fields should be treated as glossaries.'),
      '#open' => TRUE,
    ];

    $form['glossary_fields']['add'] = [
      '#type' => 'link',
      '#title' => $this->t('Add glossary field'),
      '#url' => Url::fromRoute('key_value_glossary.field_settings', ['type' => 'glossary']),
      '#attributes' => [
        'class' => ['button', 'button--primary'],
      ],
    ];

    // Display existing glossary field configurations.
    $glossary_fields = $config->get('glossary_fields') ?: [];
    if (!empty($glossary_fields)) {
      $form['glossary_fields']['existing'] = [
        '#type' => 'table',
        '#header' => [
          $this->t('Entity Type'),
          $this->t('Bundle'),
          $this->t('Field'),
          $this->t('Operations'),
        ],
        '#empty' => $this->t('No glossary fields configured yet.'),
      ];

      foreach ($glossary_fields as $delta => $field_config) {
        $form['glossary_fields']['existing'][$delta] = [
          'entity_type' => [
            '#markup' => $field_config['entity_type'],
          ],
          'bundle' => [
            '#markup' => $field_config['bundle'],
          ],
          'field_name' => [
            '#markup' => $field_config['field_name'],
          ],
          'operations' => [
            '#type' => 'operations',
            '#links' => [
              'delete' => [
                'title' => $this->t('Delete'),
                'url' => Url::fromRoute('key_value_glossary.field_settings', [
                  'type' => 'glossary',
                  'delta' => $delta,
                  'op' => 'delete',
                ]),
              ],
            ],
          ],
        ];
      }
    }

    $form['reference_fields'] = [
      '#type' => 'details',
      '#title' => $this->t('Glossary reference fields'),
      '#description' => $this->t('Select which Reference fields should be used to reference glossaries.'),
      '#open' => TRUE,
    ];

    $form['reference_fields']['add'] = [
      '#type' => 'link',
      '#title' => $this->t('Add reference field'),
      '#url' => Url::fromRoute('key_value_glossary.field_settings', ['type' => 'reference']),
      '#attributes' => [
        'class' => ['button', 'button--primary'],
      ],
    ];

    // Display existing reference field configurations.
    $reference_fields = $config->get('glossary_reference_fields') ?: [];
    if (!empty($reference_fields)) {
      $form['reference_fields']['existing'] = [
        '#type' => 'table',
        '#header' => [
          $this->t('Entity Type'),
          $this->t('Bundle'),
          $this->t('Field'),
          $this->t('Operations'),
        ],
        '#empty' => $this->t('No reference fields configured yet.'),
      ];

      foreach ($reference_fields as $delta => $field_config) {
        $form['reference_fields']['existing'][$delta] = [
          'entity_type' => [
            '#markup' => $field_config['entity_type'],
          ],
          'bundle' => [
            '#markup' => $field_config['bundle'],
          ],
          'field_name' => [
            '#markup' => $field_config['field_name'],
          ],
          'operations' => [
            '#type' => 'operations',
            '#links' => [
              'delete' => [
                'title' => $this->t('Delete'),
                'url' => Url::fromRoute('key_value_glossary.field_settings', [
                  'type' => 'reference',
                  'delta' => $delta,
                  'op' => 'delete',
                ]),
              ],
            ],
          ],
        ];
      }
    }

    $form['js_settings'] = [
      '#type' => 'details',
      '#title' => $this->t('Display settings'),
      '#description' => $this->t('Configure how glossaries are presented on the page.'),
      '#open' => TRUE,
    ];

    $form['js_settings']['highlight_all'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Highlight all terms'),
      '#default_value' => $config->get('highlight_all') ?? FALSE,
      '#description' => $this->t('If checked, all terms will be highlighted when the page loads.'),
    ];

    $default_unacceptable_elements = [
      'script',
      'style',
      'a',
      'button',
      'h1',
      'h2',
      'h3',
      'h4',
      'h5',
      'h6',
    ];

    $form['js_settings']['unacceptable_elements'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Unacceptable elements'),
      '#description' => $this->t('HTML elements that should not be processed for glossary terms. Enter one element per line without angle brackets (e.g., "div" not "&lt;div&gt;").'),
      '#default_value' => implode("\n", $config->get('unacceptable_elements') ?: $default_unacceptable_elements),
      '#rows' => 10,
    ];

    $form['js_settings']['unacceptable_containers'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Unacceptable containers'),
      '#description' => $this->t('CSS selectors for containers that should not be processed for glossary terms. Enter one selector per line.'),
      '#default_value' => implode("\n", $config->get('unacceptable_containers') ?: []),
      '#rows' => 10,
    ];

    $form['js_settings']['main_content_selector'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Main content selector'),
      '#description' => $this->t('CSS selector for the main content area where glossary terms will be processed. Default is "main".'),
      '#default_value' => $config->get('main_content_selector') ?: 'main',
    ];

    // Only show the inherit_glossary_reference option if the book module is
    // enabled.
    if ($this->moduleHandler->moduleExists('book')) {
      $form['js_settings']['inherit_glossary_reference'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Inherit glossary reference from ancestor nodes'),
        '#default_value' => $config->get('inherit_glossary_reference') ?? TRUE,
        '#description' => $this->t('When enabled, nodes will inherit the glossary reference from ancestor nodes in a Book outline. Only relevant when the book module is enabled.'),
      ];
    }

    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    // Process the unacceptable elements and containers.
    $unacceptable_elements = $form_state->getValue('unacceptable_elements');
    $unacceptable_containers = $form_state->getValue('unacceptable_containers');

    // Convert textarea values to arrays, filtering out empty lines.
    $unacceptable_elements_array = array_filter(preg_split('/\r\n|\r|\n/', $unacceptable_elements));
    $unacceptable_containers_array = array_filter(preg_split('/\r\n|\r|\n/', $unacceptable_containers));

    // Save the configuration.
    $config = $this->config('key_value_glossary.settings')
      ->set('highlight_all', $form_state->getValue('highlight_all'))
      ->set('unacceptable_elements', $unacceptable_elements_array)
      ->set('unacceptable_containers', $unacceptable_containers_array)
      ->set('main_content_selector', $form_state->getValue('main_content_selector'));

    // Save the inherit_glossary_reference setting if the book module is
    // enabled.
    if ($this->moduleHandler->moduleExists('book') && $form_state->hasValue('inherit_glossary_reference')) {
      $config->set('inherit_glossary_reference', $form_state->getValue('inherit_glossary_reference'));
    }

    $config->save();

    parent::submitForm($form, $form_state);
  }

}
