<?php

namespace Drupal\key_value_glossary;

use Drupal\book\BookManagerInterface;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\Entity\Display\EntityViewDisplayInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Field\EntityReferenceFieldItemListInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\node\NodeInterface;

/**
 * Service class for building and managing glossary functionality.
 *
 * The GlossaryBuilder handles:
 * - Checking and retrieving glossary reference fields from entities.
 * - Building glossaries from referenced entities.
 * - Managing glossary inheritance in book hierarchies.
 * - Attaching glossary data and settings to entity view builds.
 *
 * This class works with entities that have configured glossary reference fields
 * pointing to entities containing glossary fields. It supports recursive
 * glossary inheritance through book hierarchies when enabled.
 */
class GlossaryBuilder {

  /**
   * Key value glossary module simple configuration.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected ImmutableConfig $config;

  /**
   * Constructs a new GlossaryBuilder.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   The configuration factory service.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $moduleHandler
   *   The module handler service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager service.
   * @param \Drupal\book\BookManagerInterface|null $bookManager
   *   The book manager service, if the Book module is enabled.
   */
  public function __construct(
    protected ConfigFactoryInterface $configFactory,
    protected ModuleHandlerInterface $moduleHandler,
    protected EntityTypeManagerInterface $entityTypeManager,
    protected ?BookManagerInterface $bookManager = NULL,
  ) {
    $this->config = $this->configFactory->get('key_value_glossary.settings');
  }

  /**
   * Implements hook_entity_view().
   *
   * Processes entity view builds to attach glossary data and settings.
   *
   * Checks if the entity has glossary references and attaches the necessary
   * JavaScript libraries and settings for glossary term highlighting.
   *
   * @param array &$build
   *   The render array to alter.
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity being rendered.
   * @param \Drupal\Core\Entity\Display\EntityViewDisplayInterface $display
   *   The entity view display.
   * @param string $view_mode
   *   The view mode being rendered.
   */
  public function entityView(array &$build, EntityInterface $entity, EntityViewDisplayInterface $display, $view_mode): void {
    if (!$entity instanceof FieldableEntityInterface) {
      return;
    }

    if ($view_mode !== 'full') {
      return;
    }

    // We're very dependent on key_value_glossary.settings config.
    CacheableMetadata::createFromRenderArray($build)
      ->addCacheableDependency($this->config)
      ->applyTo($build);

    if (!$this->hasGlossaryReferenceField($entity)) {
      return;
    }

    if ($this->moduleHandler->moduleExists('book') &&
      $this->config->get('inherit_glossary_reference')) {
      $glossary = $this->getGlossaryRecursive($entity);
    }
    else {
      $glossary = $this->getGlossary($entity);
    }

    // Apply cacheability metadata from glossary.
    CacheableMetadata::createFromRenderArray($build)
      ->addCacheableDependency($glossary)
      ->applyTo($build);
    $glossary_data = $glossary->getMergedGlossary();

    // If we have glossary data, attach the JavaScript.
    if (!empty($glossary_data)) {
      $build['#attached']['library'][] = 'key_value_glossary/glossary';
      $build['#attached']['drupalSettings']['keyValueGlossary']['terms'] = $glossary_data;

      // Add the unacceptable elements and containers to drupalSettings.
      $build['#attached']['drupalSettings']['keyValueGlossary']['highlightAll'] = $this->config->get('highlight_all') ?? FALSE;
      $build['#attached']['drupalSettings']['keyValueGlossary']['unacceptableElements'] = $this->config->get('unacceptable_elements') ?: [];
      $build['#attached']['drupalSettings']['keyValueGlossary']['unacceptableContainers'] = $this->config->get('unacceptable_containers') ?: [];
      $build['#attached']['drupalSettings']['keyValueGlossary']['mainContentSelector'] = $this->config->get('main_content_selector') ?: 'main';
    }
  }

  /**
   * Checks if the given entity has a glossary reference field configured.
   *
   * This method checks if the entity matches any of the configured glossary
   * reference fields based on its entity type, bundle and field existence.
   *
   * @param \Drupal\Core\Entity\FieldableEntityInterface $entity
   *   The entity to check for glossary reference fields.
   *
   * @return bool
   *   TRUE if the entity has a configured glossary reference field,
   *   FALSE otherwise.
   */
  protected function hasGlossaryReferenceField(FieldableEntityInterface $entity): bool {
    try {
      $this->getGlossaryReferenceField($entity);
      return TRUE;
    }
    catch (\InvalidArgumentException $e) {
      return FALSE;
    }
  }

  /**
   * Gets the glossary reference field for a given entity.
   *
   * Searches through configured glossary reference fields to find a matching
   * field based on entity type, bundle and field name. The field must be an
   * entity reference field.
   *
   * @param \Drupal\Core\Entity\FieldableEntityInterface $entity
   *   The entity to get the glossary reference field from.
   *
   * @return \Drupal\Core\Field\EntityReferenceFieldItemListInterface
   *   The glossary reference field if found.
   *
   * @throws \InvalidArgumentException
   *   Thrown when either no matching field is found or the found field is not
   *   an entity reference field.
   */
  protected function getGlossaryReferenceField(FieldableEntityInterface $entity): EntityReferenceFieldItemListInterface {
    // Check if this entity has a glossary reference field.
    $glossary_reference_fields = $this->config->get('glossary_reference_fields') ?: [];
    foreach ($glossary_reference_fields as $field_config) {
      if ($field_config['entity_type'] === $entity->getEntityTypeId() &&
        $field_config['bundle'] === $entity->bundle() &&
        $entity->hasField($field_config['field_name'])) {
        $field = $entity->get($field_config['field_name']);
        if (!$field instanceof EntityReferenceFieldItemListInterface) {
          throw new \InvalidArgumentException('Glossary reference field is not an entity reference field.');
        }
        return $field;
      }
    }
    throw new \InvalidArgumentException('The given entity does not have a glossary reference field.');
  }

  /**
   * Gets all referenced glossary entities from given entity..
   *
   * Retrieves the glossary entities that are referenced by the entity through
   * its configured glossary reference field. If the entity does not have a
   * valid glossary reference field, returns an empty array.
   *
   * @param \Drupal\Core\Entity\FieldableEntityInterface $entity
   *   The entity containing the glossary reference field.
   *
   * @return array
   *   An array of referenced glossary entities, or an empty array if no
   *   references exist or if the entity does not have a valid glossary
   *   reference field.
   */
  protected function getReferencedGlossaryEntities(FieldableEntityInterface $entity): array {
    try {
      $field = $this->getGlossaryReferenceField($entity);
      return $field->referencedEntities();
    }
    catch (\InvalidArgumentException $e) {
      return [];
    }
  }

  /**
   * Gets the glossary field for a given entity.
   *
   * Searches through configured glossary fields to find a matching field based
   * on entity type, bundle and field name. The field must be configured in the
   * module settings as a valid glossary field.
   *
   * @param \Drupal\Core\Entity\FieldableEntityInterface $entity
   *   The entity to get the glossary field from.
   *
   * @return \Drupal\Core\Field\FieldItemListInterface
   *   The glossary field if found.
   *
   * @throws \InvalidArgumentException
   *   Thrown when no matching glossary field is found for the given entity.
   */
  protected function getGlossaryField(FieldableEntityInterface $entity): FieldItemListInterface {
    $glossary_fields = $this->config->get('glossary_fields') ?: [];
    foreach ($glossary_fields as $field_config) {
      if ($field_config['entity_type'] == $entity->getEntityTypeId() &&
        $field_config['bundle'] == $entity->bundle() &&
        $entity->hasField($field_config['field_name'])) {
        // @todo verify keyvalue field.
        return $entity->get($field_config['field_name']);
      }
    }
    throw new \InvalidArgumentException('The given entity does not have a glossary field.');
  }

  /**
   * Gets the glossary for entity and recursively from ancestors.
   *
   * For book nodes, this method recursively traverses up the book hierarchy
   * and merges glossaries from parent nodes. For non-book nodes or leaf nodes,
   * it returns only the entity's own glossary.
   *
   * @param \Drupal\Core\Entity\FieldableEntityInterface $entity
   *   The entity to get the glossary from and check for ancestors.
   *
   * @return \Drupal\key_value_glossary\MergedGlossary
   *   A MergedGlossary object containing terms from the entity and its
   *   ancestors.
   */
  protected function getGlossaryRecursive(FieldableEntityInterface $entity): MergedGlossary {
    $merged_glossary = $this->getGlossary($entity);
    // Only nodes can be books and thus have ancestors in the sense we care
    // about.
    if (!$entity instanceof NodeInterface) {
      return $merged_glossary;
    }
    // Does this entity have a parent book set?
    if (!empty($entity->book['pid'])) {
      $parent_node = $this->entityTypeManager->getStorage('node')->load($entity->book['pid']);
      if ($parent_node instanceof FieldableEntityInterface) {
        return $merged_glossary->mergeGlossary($this->getGlossaryRecursive($parent_node));
      }
    }
    return $merged_glossary;
  }

  /**
   * Gets a merged glossary from all referenced glossary entities.
   *
   * Creates a new MergedGlossary and adds terms from all referenced glossary
   * entities. For each referenced entity, gets its glossary field and merges
   * the terms into the combined glossary. Ignores any referenced entities that
   * don't have a valid glossary field configured.
   *
   * @param \Drupal\Core\Entity\FieldableEntityInterface $entity
   *   The entity containing glossary references.
   *
   * @return \Drupal\key_value_glossary\MergedGlossary
   *   A MergedGlossary object containing all terms from referenced glossaries.
   */
  protected function getGlossary(FieldableEntityInterface $entity): MergedGlossary {
    $merged_glossary = new MergedGlossary();
    $merged_glossary->addCacheableDependency($entity);
    foreach ($this->getReferencedGlossaryEntities($entity) as $glossary_entity) {
      try {
        $merged_glossary->addCacheableDependency($glossary_entity);
        $merged_glossary->mergeGlossary(
          MergedGlossary::mergedGlossaryFromGlossaryField($this->getGlossaryField($glossary_entity))
        );
      }
      catch (\InvalidArgumentException $e) {
        // Ignore exception from a glossary that doesn't have a configured
        // glossary field.
      }
    }
    return $merged_glossary;
  }

}
