<?php

namespace Drupal\key_value_glossary;

use Drupal\Core\Cache\RefinableCacheableDependencyInterface;
use Drupal\Core\Cache\RefinableCacheableDependencyTrait;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\key_value_field\Plugin\Field\FieldType\KeyValueItem;
use Drupal\key_value_field\Plugin\Field\FieldType\KeyValueLongItem;

/**
 * Manages a collection of glossary terms and their definitions.
 *
 * This class provides functionality to merge and manage glossary terms from
 * different entities, identified by their UUIDs. It maintains case-insensitive
 * term indexing and supports caching through the
 * RefinableCacheableDependencyInterface.
 */
class MergedGlossary implements RefinableCacheableDependencyInterface {
  use RefinableCacheableDependencyTrait;

  /**
   * The complete merged glossary, keyed by term.
   *
   * @var array
   */
  protected array $mergedGlossary = [];

  /**
   * Adds a term and its definition to the glossary.
   *
   * @param string $glossary_uuid
   *   The UUID of the glossary source.
   * @param string $term
   *   The term to add (case-insensitive).
   * @param string $definition
   *   The definition of the term.
   *
   * @return static
   *   Fluent API return.
   */
  public function addTerm(string $glossary_uuid, string $term, string $definition): static {
    $this->mergedGlossary[strtolower($term)][$glossary_uuid] = $definition;
    return $this;
  }

  /**
   * Gets the complete merged glossary.
   *
   * @return array
   *   An array of terms and their definitions, indexed by lowercase terms.
   */
  public function getMergedGlossary(): array {
    return $this->mergedGlossary;
  }

  /**
   * Merges another glossary into this one.
   *
   * @param \Drupal\key_value_glossary\MergedGlossary $glossary
   *   The glossary to merge into this one.
   *
   * @return static
   *   Returns this object for method chaining.
   */
  public function mergeGlossary(MergedGlossary $glossary): static {
    $this->mergedGlossary = array_merge_recursive($this->mergedGlossary, $glossary->getMergedGlossary());
    $this->addCacheableDependency($glossary);
    return $this;
  }

  /**
   * Creates a merged glossary from a field item list.
   *
   * @param \Drupal\Core\Field\FieldItemListInterface $field_items
   *   The field items containing glossary terms and definitions.
   *
   * @return static
   *   A new MergedGlossary instance containing the field items.
   */
  public static function mergedGlossaryFromGlossaryField(FieldItemListInterface $field_items): static {
    $merged_glossary = new static();
    foreach ($field_items as $item) {
      if (!$item instanceof KeyValueItem && !$item instanceof KeyValueLongItem) {
        continue;
      }
      $merged_glossary->addTerm($field_items->getEntity()->uuid(), $item->key, $item->value);
    }
    return $merged_glossary;
  }

}
